/*  File:  art.c  */

/*  Adaptive Resonance Theory 1
 *
 *  James A. Freeman and David M. Skapura
 *  Neural Networks
 *  Algorithms, Aplications, and Programming Techniques
 *
 *  8.2.3: ART1 Processing Summary
 */

#ifdef __MSDOS__
#ifndef __COMPACT__
#error Memory model COMPACT required
#endif  /* __COMPACT__ */
#include <conio.h>
#include <dir.h>
#endif  /* __MSDOS__ */

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#define BUFSIZE 2048

enum BOOL { FALSE = 0, TRUE = 1 };

struct F2
{
    float
        T,
        *up,
        *dn;
};

struct VEC
{
    char
	*name;
    int
	*v;
};

int
    breed,
    hoog,
    scr_width,
    veclen,
    n_vec = 0,
    max_vec = 0,
    n_f2 = 0,
    max_f2 = 0,
    n_used = 0,
    input_line;

float
    A1,
    B1,
    C1,
    D1,
    L,
    rho = 0,
    *S,
    *V;

struct VEC
    *vec = NULL;

struct F2
    *f2 = NULL;

char
    *programname,
    buffer  [BUFSIZE + 1],
    *no_mem_buffer,
    *infile = NULL,
    *no_mem_buffer,
    Out_of_memory [] = "Out of memory";

void
    show (int v),
    get_programname (char const *argv0),
    process_args (int argc, char *argv []),
    setup (void),
    make_f2_unit (void),
    read_file (void),
    syntax (void),
    errit (char const *format, ...),
    *s_malloc (size_t size),
    *s_realloc (void *block, size_t size);
char
    *s_strdup (char const *s),
    *getline (FILE *fp, enum BOOL required, char const *filename);
char const
    *get_arg (int argc, char *argv [], int *index);
float
    *alloc_vector_float (int i1, float value);

int main (int argc, char *argv [])
{
    int
	i,
	j,
	v,
        coded,
	f2win;
    float
	max,
	sum,
        nwup,
        nwdn,
	S_mag,
	I_mag;

    no_mem_buffer = (char *) malloc (1024);

    get_programname (argv [0]);

    process_args (argc, argv);

    read_file ();

    setup ();

    coded = 0;
    v = -1;
    while (coded++ < n_vec) {

        if (++v == n_vec)
            v = 0;

        printf ("%s:", vec [v].name);

        if (n_used == n_f2)
            make_f2_unit ();

        I_mag = 0.0;
        for (i = 0; i < veclen; i++) {
            S [i] =(((float) vec [v].v [i]) /
                    (1.0 + A1 * (((float) vec [v].v [i]) + B1) + C1)
                   > 0.0) ? 1.0 : 0.0;
            I_mag += vec [v].v [i] ? 1.0 : 0.0;
        }

        for (j = 0; j < n_f2; j++) {
            sum = 0.0;
            for (i = 0; i < veclen; i++)
                sum += S [i] * f2 [j].up [i];
            f2 [j].T = sum;
        }

        for (;;) {

            max = -MAXFLOAT;
            for (i = 0; i < n_f2; i++) {
		if (f2 [i].T > max) {
                    max = f2 [i].T;
                    f2win = i;
                }
            }
            if (max <= 0.0)
                errit ("Really weird!");

            S_mag = 0.0;
            for (i = 0; i < veclen; i++) {
		V [i] = f2 [f2win].dn [i];
                S_mag += (S [i] = ((((float) vec [v].v [i]) + D1 * V [i] - B1)
                                 / (1.0 +
                                    A1 * (((float) vec [v].v [i]) + D1 * V [i])
                                  + C1) > 0.0) ? 1.0 : 0.0);
            }

            printf (" %i", f2win + 1);

            if (S_mag / I_mag < rho)
                f2 [f2win].T = -MAXFLOAT;
            else
                break;
        }
        printf ("\n");

        /* ongebruikte unit? */
        if (f2win >= n_used)
            n_used = f2win + 1;

        for (i = 0; i < veclen; i++) {
            if (S [i] > 0.5) {
                nwup = L / (L - 1.0 + S_mag);
                nwdn = 1.0;
            } else
                nwup = nwdn = 0.0;
	    if (nwup != f2 [f2win].up [i]) {
		coded = 0;
                f2 [f2win].up [i] = nwup;
            }
	    if (nwdn != f2 [f2win].dn [i]) {
		coded = 0;
		f2 [f2win].dn [i] = nwdn;
            }
        }

        if (coded == 0)
            show (v);
    }

    printf ("\nrho: %f\n\n", rho);

    return 0;
}

void show (int nr)
{
    int
	i,
	j,
	k,
	v,
	line,
	lines,
	v_per_line;
    v_per_line = (scr_width - breed - 5) / (breed + 2);
    lines = n_used / v_per_line;
    if (lines * v_per_line < n_used)
	lines++;
    for (line = 0; line < lines; line++) {
	for (j = 0; j < hoog; j++) {
	    if (line == 0) {
		for (k = 0; k < breed; k++)
		    printf (vec [nr].v [j * breed + k] ? "@" : ".");
		printf (" :  ");
	    } else
		printf ("%*s    ", breed, "");
	    for (i = 0; i < v_per_line; i++) {
		v = line * v_per_line + i;
		if (v < n_used) {
		    for (k = 0; k < breed; k++)
			printf ((f2 [v].dn [k + breed * j] > 0.5) ? "@" : ".");
		    printf ("  ");
		}
	    }
	    printf ("\n");
        }
	printf ("\n");
    }
}

void setup ()
{
#ifdef __MSDOS__
    struct text_info
	ti;
    gettextinfo (&ti);
    scr_width = ti.screenwidth;
#else
    char
        *c;
    c = getenv ("COLUMNS");
    if (c) {
        scr_width  = atoi (c);
        if (scr_width < 1)
            scr_width = 80;
    } else
        scr_width = 80;
#endif

    A1 =  1.0;
    B1 =  1.5;
    C1 =  5.0;
    D1 =  0.9;
    L  =  3.0;

    if (rho <= 0.0)
        rho = 0.5;

    S = alloc_vector_float (veclen, 0.0);
    V = alloc_vector_float (veclen, 0.0);
}

void make_f2_unit ()
{
    while (n_f2 >= max_f2) {
        max_f2 += 64;
        f2 = (struct F2 *) s_realloc (f2, max_f2 * sizeof (struct F2));
    }
    f2 [n_f2].up   = alloc_vector_float
                        (veclen, 0.8 * (L / (L - 1.0 + (float) veclen)));
    f2 [n_f2++].dn = alloc_vector_float
                        (veclen, 1.4 * ((B1 - 1.0) / D1));
}

void read_file ()
{
    int
	i,
	j,
	k;
    char const
	*file_error = "Fout in bestand \"%s\", regel %i",
	*illegal = "Ongeldige waarde voor %s";

    FILE
	*fp;

    if (! infile)
	errit ("Naam invoerbestand ontbreekt");

    fp = fopen (infile, "r");
    if (! fp)
	errit ("Bestand \"%s\" niet gevonden", infile);
    input_line = 0;

    getline (fp, TRUE, infile);
    if (sscanf (buffer, "%i %i", &breed, &hoog) != 2)
	errit (file_error, infile, input_line);
    if (breed < 1)
	errit (illegal, "breedte");
    if (hoog < 1)
	errit (illegal, "hoogte");

    veclen = breed * hoog;

    while (getline (fp, FALSE, NULL)) {
	while (n_vec >= max_vec) {
	    max_vec += 64;
	    vec = (struct VEC *)
		s_realloc (vec, max_vec * sizeof (struct VEC));
	}
	vec [n_vec].name = s_strdup (buffer);
	vec [n_vec].v = (int *) s_malloc (veclen * sizeof (int));
	for (i = 0; i < hoog; i++) {
	    getline (fp, TRUE, infile);
	    j = 0;
	    k = -1;
	    while (j < breed) {
		switch (buffer [++k]) {
		    case ' ':
		    case '\t':
			break;
                    case '-':
		    case '0':
			vec [n_vec].v[i * breed + j++] = 0;
			break;
                    case '+':
		    case '1':
			vec [n_vec].v[i * breed + j++] = 1;
			break;
		    default:
			errit (
			    file_error,
			    infile,
			    input_line
			);
		}
	    }
	}
	n_vec++;
    }
    fclose (fp);
}

char *getline (FILE *fp, enum BOOL required, char const *filename)
{
    int
	i;

    for (;;) {
	if (fgets (buffer, BUFSIZE, fp) == NULL) {
	    if (required)
		errit ("Unexpected end of file in \"%s\"", filename);
	    else
		return NULL;
	}
	input_line++;
	i = strlen (buffer);
	while (i && isspace (buffer [i - 1]))
	    buffer [--i] = '\0';
	i = 0;
	while (buffer [i] && isspace (buffer [i]))
	    i++;
	if (buffer [i] == '#')
	    continue;
	if (buffer [i]) {
	    memmove (buffer, buffer + i, strlen (buffer) + 1);
	    return buffer;
	}
    }
}

void syntax ()
{
    fprintf (
	stderr,
	"\nAdaptive Resonance Theory 1\n\n"
        "(c) P. Kleiweg 1996\n\n"
	"Syntax: %s -i string [-r float]\n\n"
	"-i : inputfile\n"
	"-r : vigilance: 0.0 < rho <= 1.0\n\n",
	programname
    );
    exit (1);
}

void process_args (int argc, char *argv [])
{
    int
	i;

    if (argc == 1)
	syntax ();

    for (i = 1; i < argc; i++) {
	if (argv [i][0] != '-' && argv [i][0] != '/')
	    errit ("Illegal argument");
	switch (argv [i][1]) {
	    case 'i':
		infile = s_strdup (get_arg (argc, argv, &i));
		break;
            case 'r':
                rho = atof (get_arg (argc, argv, &i));
		if (rho <= 0.0 || rho > 1.0)
                    errit ("Illegal value for rho");
		break;
	    default:
		errit ("Illegal option '%s'", argv [i]);
	}
    }
}

char const *get_arg (int argc, char *argv [], int *index)
{
    if (argv [*index][2])
	return argv [*index] + 2;

    if (*index == argc - 1)
	errit ("Argument voor optie '%s' ontbreekt", argv [*index]);

    return argv [++*index];
}

void get_programname (char const *argv0)
{
#ifdef __MSDOS__
    char
	name [MAXFILE];
    fnsplit (argv0, NULL, NULL, name, NULL);
    programname = s_strdup (name);
#else   /* unix */
    char
	*p;
    p = strrchr (argv0, '/');
    if (p)
	programname = s_strdup (p + 1);
    else
	programname = s_strdup (argv0);
#endif
}

float *alloc_vector_float (int i1, float value)
{
    int
	i;
    float
	*f;

    f = (float *) s_malloc (i1 * sizeof (float));
    for (i = 0; i < i1; i++)
	f [i] = value;
    return f;
}

void *s_malloc (size_t size)
{
    void
	*p;

    p = malloc (size);
    if (! p) {
	free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void *s_realloc (void *block, size_t size)
{
    void
	*p;

    p = realloc (block, size);
    if (! p) {
	free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

char *s_strdup (char const *s)
{
    char
	*s1;

    if (s) {
	s1 = (char *) s_malloc (strlen (s) + 1);
	strcpy (s1, s);
    } else {
	s1 = (char *) s_malloc (1);
	s1 [0] = '\0';
    }
    return s1;
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}
